/*! @file
 * @brief
 * mpqs sieving stuff that needs to be included at compile-time
 */


void do_sieving_Squares()
{
  // Quadrate von statischen (und ggf. dynamischen Faktoren) sieben.
  // Wichtig: Daher nur den einfachen Logarithmus des Faktors subtrahieren!
  // Bemerkung: Zur Effizienzsteigerung wird davon ausgegangen, dass jedes Quadrat nur einmal
  //            innerhalb des gesamten Siebintervalls trifft!

  //cout << "Quadrate sieben Start" << endl;
  if (Sieve_Delta_HeapOfSquares.empty()) return; // es gibt nichts zu sieben...
  TSieve_Delta sieb_delta = Sieve_Delta_HeapOfSquares.top();
  register int delta = sieb_delta.delta-SieveOffset;
  
  while (delta<PhysicalSieveSize)
    {
      Sieve_Delta_HeapOfSquares.pop(); // Element aus der Priorittswarteschlange nehmen
      SieveArray[delta]-=static_cast<TSieveElement>(ceil(log_Multiplikator*log(sieb_delta.factor))); // Teiler im Sieb vermerken
      if (Sieve_Delta_HeapOfSquares.empty()) break;
      sieb_delta = Sieve_Delta_HeapOfSquares.top();
      delta = sieb_delta.delta-SieveOffset;
    }
  //cout << "Quadrate sieben Ende" << endl;
}


void do_sieving_DynamicFactors()
{
  if (Sieve_Delta_Heap.empty()) return; // es gibt nichts zu sieben...
  short int HitCount=0;
  TSieve_Delta sieb_delta = Sieve_Delta_Heap.top();
  register int delta = sieb_delta.delta-SieveOffset;
  goto Schleifentest;
  
 Schleifenanfang_weiterer_Faktor:
  ++HitCount;
  //if (HitCount>=MaxHits) { cerr << "MaxHits erhhen!"; exit(1); } 
 Schleifenanfang_normal:
  Sieve_Delta_Heap.pop(); // Element aus der Priorittswarteschlange nehmen
  SieveArray[delta]-=log_DynamicLargePrime(sieb_delta.factor); // Teiler im Sieb vermerken
  
  // Da dynamische Faktoren selten auftreten, diese Faktoren schon whrend
  // des Siebens vermerken...
  // ...das erhht die Effizienz bei der Refaktorisierung...    
  Hits[HitCount].Faktor=sieb_delta.factor;

#ifdef USE_FAKEHEAP
  if (Sieve_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben?
#else
  sieb_delta.delta += sieb_delta.factor; // nchste Hit-Position
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
   {
     Sieve_Delta_Heap.push(sieb_delta); // wieder in Priorittswarteschlange eintragen
   }
  else
    if (Sieve_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben?
#endif
  
  sieb_delta = Sieve_Delta_Heap.top(); // nchsten Eintrag aus Priorittswarteschlange holen
  register int delta2; delta2 = sieb_delta.delta-SieveOffset;
  if (delta==delta2) goto Schleifenanfang_weiterer_Faktor;  
  if (SieveArray[delta]<0) // Relation gefunden?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      StaticRelations::insert(new CRelation(delta+SieveOffset, HitCount+1)); /* dann einfgen */
    }
  delta=delta2;
  
  HitCount=0; // und auf ein neues
 Schleifentest:
  if (delta<PhysicalSieveSize) goto Schleifenanfang_normal; // noch im aktuellen (physikalischen) Intervall?
  return; // fertig mit diesem Intervall
  
 Schlange_leer:
  if (SieveArray[delta]<0) // Relation gefunden?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      StaticRelations::insert(new CRelation(delta+SieveOffset,HitCount+1)); // dann einfgen
    }
  return; 
}


#if 0
#include <map>
void Werteverteilung_im_Sieb(void)
{
  // fr statistische Zwecke kann es ganz sinnvoll sein, wenn man
  // sich mal anschauen kann, wie die Werte um den Nullwert herum verteilt sind...
  typedef map<TSieveElement,unsigned int> TStats;
  static TStats S;
  for (int i=0; i<PhysicalSieveSize; ++i) S[SieveArray[i]]++;

  static int z = 1000;
  if (--z==0)
   {
     z=1000;

     for (TStats::iterator pos=S.begin(); pos!=S.end(); ++pos)
      cout << static_cast<signed long int>(pos->first) << " : " << pos->second << endl;

     cout << "Eingabe fr weiter" << endl;
     string s; cin >> s;
   }
}
#endif


void recompute_all_Deltas()
{
  /* Beim Wechsel auf ein neues Polynom werden alle bisherigen Deltawerte der Primzahlen
     ungltig. Daher ist es erforderlich, diese neu zu berechnen */

  //cout << "Berechne die Deltas neu" << endl;

  // zunchst alle (mglicherweise) noch bestehenden dynamischen Faktoren aus der Warteschlange nehmen:
#ifdef USE_FAKEHEAP
//  cout << "clearing Fakeheap with " << Sieve_Delta_Heap.size() << " elements..." << endl;
  Sieve_Delta_Heap.clear();
#else
  while (!Sieve_Delta_Heap.empty()) Sieve_Delta_Heap.pop(); // fuer priority_queue, die kein clear hat
#endif

  // ebenso die Quadrate:
  //Sieve_Delta_HeapOfSquares.clear();
  while (!Sieve_Delta_HeapOfSquares.empty()) Sieve_Delta_HeapOfSquares.pop(); // fuer priority_queue, die kein clear hat
 
  
  // als erstes die Deltas aus der statischen Faktorbasis ermitteln
  unsigned int inv_A2_modp, PolyB;
  mpz_t inv_A2, P, wuq, x, y;
  mpz_init(inv_A2); mpz_init(P); mpz_init(wuq); mpz_init(x); mpz_init(y);


#if 1 /* Test!!!! 2004-03-09 */
  for (int i=FB_maxQuadrate+1; i<StaticFactorbase::Size() && PrimeNumbers[i]<25000; ++i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
	  // Fr alle/einige "groe" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) parallel mit den dynamischen Faktoren sieben.
          const unsigned int P = PrimeNumbers[i]*PrimeNumbers[i];
          const unsigned int wuq = SQRT_kN_of_PrimeSquares[i];
          inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),P),P);
          PolyB=mpz_remainder_ui(Polynom.get_B(),P);
      
          // Delta1
          signed int d = wuq>=PolyB ? wuq-PolyB : wuq-PolyB+P; // wg. berlauf
          d=mulmod(d,inv_A2_modp,P);
          d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }

	  // Delta2
          unsigned int h = wuq+PolyB; d=P-mulmod(h,inv_A2_modp,P); d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif



#ifdef SIEVING_LARGE_SQUARES
  for (int i=StaticFactorbase::Size()-1; i>FB_maxQuadrate && PrimeNumbers[i]>25000; --i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
	  // Fr alle "groen" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) zusammen mit den dynamischen Faktoren sieben.
	  mpz_set_ui(P,PrimeNumbers[i]); mpz_mul(P,P,P); // Quadrat der Primzahl holen
	  mpz_invert(inv_A2,Polynom.get_A2(),P);

	  // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
	  // Wurzeln von kN modulo Primzahl^2 ermitteln
	  const int Wu = SQRT_kN_of_PrimeNumbers[i];
	  mpz_set_ui(y,2*Wu);
	  if (mpz_invert(y,y,P)==0)
	    {
	      cerr << "PQ-root for " << PrimeNumbers[i] << ": inverse doesn't exist!" << endl;
	      exit(1);
	    }
	  
	  mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
	  mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG	  
	  // ggf. Prfung, ob Wurzel korrekt ist:
          mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
          if (mpz_cmp_ui(x,0)!=0)
           {
             cerr << "PQ-root incorrect!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[i] << endl;
             cerr << "kN= " << kN << endl;
             cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[i] << endl;
             cerr << "SQRT(kN) mod p^2=" << wuq << endl;
             cerr << "but we got (wrong value): " << x << endl;
             exit(1); 
           }
#endif
	  // nun die Deltas:

	  // Delta1
	  mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }
	  
	  // Delta2
	  mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif

  
  // Faktorbasis: erste Siebtreffer berechnen; Primzahl 2 berspringen, da Polynom damit Probleme hat...
  for (int i=StaticFactorbase::Size()-1; PrimeNumbers[i]>2; --i)
    {
      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimeNumbers[i]),PrimeNumbers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimeNumbers[i]);
      
      register int d;
      d=SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][0]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);
      
      d=PrimeNumbers[i]-SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][1]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);

#ifdef DEBUG
      //ggf. testen, ob Werte in Ordnung sind...
      Polynom.get_values(Delta_of_PrimeNumbers[i][0],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0) 
	{ 
	  cerr << "Deltawert0 " << Delta_of_PrimeNumbers[i][0] << " liefert Teiler fr "
	       << PrimeNumbers[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_of_PrimeNumbers[i][0]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][0]<0) Delta_of_PrimeNumbers[i][0]+=PrimeNumbers[i];
	}
      Polynom.get_values(Delta_of_PrimeNumbers[i][1],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0)
	{ 
	  cerr << "Deltawert1 " << Delta_of_PrimeNumbers[i][1] << " liefert Teiler fr "
	       << PrimeNumbers[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_of_PrimeNumbers[i][1]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][1]<0) Delta_of_PrimeNumbers[i][1]+=PrimeNumbers[i];
	}
#endif
    }

#ifdef SIEVING_MORE_LARGE_SQUARES
  // und weitere Quadrate auerhalb der statischen Faktorbasis pushen:
  {
    unsigned int Primzahl = biggest_Prime_in_Factorbase;
    const unsigned int Obergrenze = 1000000; // or maybe 4*biggest_Prime_in_Factorbase;
    cout << "pushing more prime squares up to " << Obergrenze << endl; 
    while (Primzahl<Obergrenze)
     {
       do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; 
           while( Primzahl%3==0 || Primzahl%5==0 || Primzahl%7==0
                  || Primzahl%11==0 || Primzahl%13==0 || Primzahl%17==0
                  || !probab_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(P,Primzahl);
	}
       while (mpz_legendre(kN,P)<0); /* Primzahl gltig fr Primzahlbasis? */

       mpz_mul(P,P,P); // Quadrat der Primzahl holen
       mpz_invert(inv_A2,Polynom.get_A2(),P);

       // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
       // Wurzeln von kN modulo Primzahl^2 ermitteln
       const int Wu = SQRT_kN_mod_PrimeNumber(Primzahl);
       mpz_set_ui(y,2*Wu);
       if (mpz_invert(y,y,P)==0)
        {
          cerr << "PQ-root for " << Primzahl << ": inverse doesn't exist!" << endl;
          exit(1);
	}
	  
       mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
       mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG
       // ggf. Prfung, ob Wurzel korrekt ist:
       mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
       if (mpz_cmp_ui(x,0)!=0)
        {
          cerr << "PQ-root incorrect!" << endl;
          cerr << "Primzahl=" << Primzahl << endl;
          cerr << "kN= " << kN << endl;
          cerr << "SQRT(kN) mod p=" << Wu << endl;
          cerr << "SQRT(kN) mod p^2=" << wuq << endl;
          cerr << "but we got (wrong value): " << x << endl;
          exit(1); 
        }
#endif
       // nun die Deltas:

       // Delta1
       mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
          // Bemerkung: Die Primzahl wird im Sieb_Quadrate_Heap *und* 
          // zustzlich auch im Sieve_Delta_Heap abgelegt, wenn es
          // sich *nicht* um einen (bereits bekannten) dynamischen Faktor
          // handelt (denn dieser wird/wurde ja bereits dort vermerkt und
          // eine Doppelzhlung sollte vermieden werden).
          // Damit ist sichergestellt, dass
          //  - der Wert in den integer-Bereich pat
          //  - der Wert doppelt (d.h. als Quadrat) im Sieb vermerkt wird.
          // Der Effekt eines zu hufigen Siebens im Falle, dass
          // die Primzahl als dynamischer Faktor whrend des Siebens entdeckt
          // und dann nachtrglich eingefgt werden knnte, wird vernachlssigt.
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // und ablegen
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta1)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta1)" << endl; };
#endif
        }

       // Delta2
       mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
          // siehe auch Bemerkung fr Delta1
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // und ablegen
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta2)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta2)" << endl; };
#endif
        }

     }
  }
#endif

#if defined(VERBOSE) || defined(DEBUG)
  if (!Sieve_Delta_HeapOfSquares.empty())
   cout << Sieve_Delta_HeapOfSquares.size() << " squares have been pushed." << endl;
#endif
  
  for (int i=0; i<NumberOf_more_PrimePowers; ++i)
    {
      // Fr diese Zahlen auch die Deltas bzgl. der jeweiligen Potenz ermitteln
      // (fr das Sieben zusammen mit den statischen Faktoren)

      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimePowers[i]),PrimePowers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimePowers[i]);

      register int d;
      d=SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][0]=mulmod(d,inv_A2_modp,PrimePowers[i]);
      
      d=PrimePowers[i]-SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][1]=mulmod(d,inv_A2_modp,PrimePowers[i]);
    }
  
  mpz_clear(inv_A2); mpz_clear(P); mpz_clear(wuq); mpz_clear(x); mpz_clear(y);
  
  // dann die dynamische Faktorbasis behandeln:
  TDynamicFactorRelations::iterator pos;
  for (pos=DynamicFactorRelations.begin();
       pos!=DynamicFactorRelations.end() && pos->sieveable(); ++pos) compute_Deltas_for(*pos, -LogicalSieveSize);
  
 // cout << "Deltaberechnung beendet." << endl;
}
