/*! @file
 * @brief
 * implementation of class SpecialRelations; handling double large prime stuff
 */


#include "SpecialRelations.H"

SpecialRelations::TSpecialFactorRelations SpecialRelations::SpecialFactorRelations;
filebuf SpecialRelations::FileBuffer;
ostream SpecialRelations::SpecialRelations_to_file(&SpecialRelations::FileBuffer);
istream SpecialRelations::SpecialRelations_from_file(&SpecialRelations::FileBuffer);


void SpecialRelations::Load()
{
  // Special Relations einlesen
  // zustzlich Speicherkompaktifizierung der special relations
  // auf dem File durchfhren, d.h. ungltige (=als redundant
  // markierte) Relationen streichen.
  cout << "Reading special relations and reorganizing file..." << endl;
  unsigned int u_hits = 0, g_hits = 0;
  streampos rpos=0, wpos=0;
  // Special Relations einlesen
  SpecialRelations_from_file.seekg(0,ios::beg);
  SpecialRelations_to_file.seekp(0,ios::beg);

  while (SpecialRelations_from_file.peek()=='G')
   {
     TSpecialFactorRelation relation;
     relation.fpos = SpecialRelations_from_file.tellg();
     string s; SpecialRelations_from_file >> s; // "G " einlesen
     SpecialRelations_from_file >> relation.factor; // special relation factor einlesen
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
     ++g_hits;
    }
  wpos=rpos=SpecialRelations_from_file.tellg(); //Schreibposition=Leseposition

  while (true) 
    {
      SpecialRelations_from_file.seekg(rpos);
      while (SpecialRelations_from_file.peek()=='U')
       {
         SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
         ++u_hits;
       }

      int cached_g = 0;
      const int max_cached_g = 1000;
      string sG[max_cached_g];
      while (SpecialRelations_from_file.peek()=='G')
       {
         while (SpecialRelations_from_file.peek()!='\n')
          {
            char buffer[2048];
            SpecialRelations_from_file.get(buffer, sizeof(buffer), '\n');
            sG[cached_g]+=buffer;
          }
         SpecialRelations_from_file.ignore(10,'\n');
         if (++cached_g>=max_cached_g) break;
         while (SpecialRelations_from_file.peek()=='U')
          {
            SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
            ++u_hits;
          }
       }
      rpos=SpecialRelations_from_file.tellg();
      if (cached_g==0) break;
      SpecialRelations_to_file.seekp(wpos);

      for (int i=0; i<cached_g; ++i)
       {
         TSpecialFactorRelation relation;
         relation.fpos = SpecialRelations_to_file.tellp();
         {
           istringstream is(sG[i]);
           string s; is >> s;
           is >> relation.factor;
         }
         SpecialRelations_to_file << sG[i] << "\n";
         SpecialFactorRelations.insert(relation);
         relation.factor.swap();
         SpecialFactorRelations.insert(relation);
         relation.factor.swap();
       };
      SpecialRelations_to_file.flush();
      wpos=SpecialRelations_to_file.tellp();
      g_hits+=cached_g;
      cout << g_hits << " special relations written, "
           << u_hits << " removed." << "\r" << flush;
      //cout << (long long int)static_cast<streamoff>(rpos) << " , " << (long long int)static_cast<streamoff>(wpos) << endl;
    }
   cout << endl;
   
  // mangels truncate-Funktion den Rest mit U's auffllen
  {
    streampos pos, ende;
    pos = wpos;
    SpecialRelations_to_file.seekp(0, ios::end);
    ende = SpecialRelations_to_file.tellp();
    SpecialRelations_to_file.seekp(wpos);
    //cout << (long long int)static_cast<streamoff>(pos) << " , " << (long long int)static_cast<streamoff>(ende) << endl;

    char buffer[4096];
    for (unsigned int i=0; i<sizeof(buffer); ++i) buffer[i]='U';
    while (static_cast<streamoff>(SpecialRelations_to_file.tellp()) < static_cast<streamoff>(ende))
      {
        SpecialRelations_to_file.write(buffer,sizeof(buffer));
        //cout << (long long int)static_cast<streamoff>(SpecialRelations_to_file.tellp()) << " , " << (long long int)static_cast<streamoff>(ende) << endl;
      }
    SpecialRelations_to_file << flush;
    SpecialRelations_to_file.seekp(pos);
  }
  SpecialRelations_from_file.clear();
  cout << "Recover: " << SpecialFactorRelations.size()
       << " special relations read, " << endl
       << u_hits << " invalid DLP-relations eliminated." << endl;
}



#ifndef IS_CLIENT

void SpecialRelations::CycleSearch()
{
  // Wenn man einen Zyklus von Large-Primes durch Double-Large-Primes
  // erzeugen kann, so neutralisieren sich die beteiligten Large-Primes und
  // man erhlt eine vollstndig in der statischen Faktorbasis aufgehende
  // Relation.

  if (SpecialFactorRelations.empty()) return; // es gibt nichts zu finden
  cout << "Starting cycle find for Special-Factor-Set" << endl;

  TSpecialFactorRelations arbeitsmenge;
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;

  arbeitsmenge.swap(SpecialFactorRelations);
  // Hinweis: um nicht doppelten Speicherbedarf zu haben, werden die
  // Special-Factor-Relations in die Arbeitsmenge geschoben (und nicht kopiert)
  // es ist dann darauf zu achten, dass sie von der Arbeitsmenge auch wieder
  // zurckgeschoben (und nicht nur gelscht) werden...
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;

  intset knotenmenge;
  typedef list<TSpecialFactorRelation> Tkantenliste;
  Tkantenliste kantenliste;

  TSpecialFactorRelations::iterator pos;
  TSpecialFactorRelation SearchRelation;
  int rechts,links;

  while (!arbeitsmenge.empty())
   {
     knotenmenge.clear(); // wichtig! (wegen Schleifenwiederholungen)
     pos=arbeitsmenge.begin();
     TSpecialFactorRelation relation = *pos;
     kantenliste.push_back(relation);
     links=relation.factor.LP1();
     rechts=relation.factor.LP2();
     knotenmenge.insert(links);
     knotenmenge.insert(rechts);
     arbeitsmenge.erase(pos); SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();

     //cout << "Starte neue Runde der Zyklensuche." << endl;
     while (!kantenliste.empty())
      {
        //for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
        // cout << (*p).factor << "|";
        //cout << endl;
        //cout << "Suche nach " << links << endl;
	//cout << "Arbeitsmenge enthlt: " << endl;
        //for (pos=arbeitsmenge.begin(); pos!=arbeitsmenge.end(); ++pos)
        // cout << (*pos).factor << ",";
	//cout << endl;

        SearchRelation.factor.set_for_search(links);
        pos=arbeitsmenge.lower_bound(SearchRelation);
        if (pos!=arbeitsmenge.end() && (*pos).factor.DLP_divisible_by(links) )
         {
           // links kann eine Kante eingefgt werden
           relation = *pos;
           arbeitsmenge.erase(pos); // alten Specialfactor entfernen
           SpecialFactorRelations.insert(relation);
           relation.factor.swap();
           arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
           SpecialFactorRelations.insert(relation);
           relation.factor.swap();
           links=relation.factor/links;
           if (knotenmenge.find(links)!=knotenmenge.end())
            { 
              // Zyklus gefunden!
	      statistical_data::gefundene_DLP_Zyklen++;
	      cout << "DLP-cycle found... " << endl;
              // nun noch: 
              // Zyklus auswerten und erhaltene Relation ins Gleichungssystem
              // einfgen.
              CRelation *GL = new CRelation;
              GL->combine(SpecialRelations_from_file,relation.fpos);
              mpz_mul_ui(GL->Delta,GL->Delta,links);
	      mpz_mod(GL->Delta,GL->Delta,n);
              rechts=relation.factor/links;
	      //cout << relation.factor;

              // Kantenliste bis zu der Position, wo der Wert "links"
              // erstmalig enthalten ist, durchgehen -> dort endet der Zyklus.
	      // dabei noch die Wurzel des Produkts der Specialfaktoren
              // hinzumultiplizieren (beim Zyklus sind dies alle beteiligten
              // Large Primes in einfacher Potenz), also die Werte "rechts"
	      unsigned int Zyklenlaenge=1;
              for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
	       {
                 Zyklenlaenge++;
		 //cout << " - " << p->factor;
                 mpz_mul_ui(GL->Delta,GL->Delta,rechts);
	         mpz_mod(GL->Delta,GL->Delta,n);
                 GL->combine(SpecialRelations_from_file,(*p).fpos);
                 if ((*p).factor.DLP_divisible_by(links)) break;
		 rechts=(*p).factor/rechts;
               }
              //cout << endl;
              cout << "DLP-cycle size: " << Zyklenlaenge << endl;
	      GL->SanityCheck(); // aus Sicherheitsgrnden...
     	      StaticRelations::insert(GL);
	    
  	      // Anschlieend eine verantwortliche (wir whlen die letzte Kante)
              // aus der Special-Relation-Menge entfernen:
              SpecialFactorRelations.erase(relation); // alten Specialfactor entfernen
              relation.factor.swap();
              SpecialFactorRelations.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
              relation.factor.swap();
              {
	        // nicht vergessen, die Relation auf File zu markieren!
                const streampos pos = SpecialRelations_to_file.tellp(); // Schreibposition merken
                SpecialRelations_to_file.seekp(relation.fpos); SpecialRelations_to_file << "U " << flush; // ungltig markieren
                SpecialRelations_to_file.seekp(pos); // wieder an die alte position
              }
 
              // Schlielich: Zyklensuche fortsetzen.
	      links=relation.factor/links;
              continue;
            }
           else
            {
              // kein Zyklus gefunden -> Kante und Knoten einfgen
              knotenmenge.insert(links);
              kantenliste.push_front(relation);
	      //cout << "Fge Kante " << relation.factor << " hinzu " << endl;
              continue;
            } 
         }
        // keine Kante zum linksseitigen Einfgen gefunden
        // -> Alternative 1: rechts einfgen (zum Zyklusfinden nicht notwendig)
        // -> Alternative 2: linksseitig eine Kante entfernen
	//cout << "Entferne Kante " << kantenliste.begin()->factor << endl;
        links=kantenliste.begin()->factor/links;
        kantenliste.pop_front(); // also linksseitig eine Kante entfernen
      }
   }
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;
  cout << "cycle find finished." << endl;
}

bool SpecialRelations::insert(const CmpqsFactor &DLP, CRelation *GL, const short int HitCount /* =0 */)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren,
  // so dass man GL anschlieend ins Gleichungssystem eintragen kann.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat!" << endl;
     statistical_data::Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  const TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  if (pos==SpecialFactorRelations.end())
   {
     relation.fpos=GL->save(SpecialRelations_to_file,relation.factor,HitCount);
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
        SpecialRelations::CycleSearch();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit by already known Special Factor!" << endl;
     statistical_data::Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.assign_to_mpz(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->combine(SpecialRelations_from_file,(*pos).fpos);
     //StaticRelations::insert(GL); // sollte an anderer Stelle gemacht werden!
     return false;
   }
}


bool SpecialRelations::SpecialFactor_splitable(const CmpqsFactor &DLP)
{
  // Testen, ob Special-Factor erfolgreich gesplittet werden kann
  // true -> ja, false -> nein
  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     // cout << "Special case: DLP is a square!" << endl;
     return true;
   }
  TSpecialFactorRelation relation;
  relation.factor=DLP;
  TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  return (pos!=SpecialFactorRelations.end());
}

void SpecialRelations::split_by_primefactor(const int Primfaktor)
{
  // Specialfaktoren liegen auerhalb der Primzahlbasis.
  // Mit ihnen wird nicht gesiebt. Sie sind auerdem zusammengesetzt.
  // Wenn es gelingt, von einem Specialfaktor einen Primfaktor aus der dynamischen FB
  // abzusplitten, dann bestehen gute Chancen, dadurch einen neuen dynamischen Faktor
  // zu erhalten oder gar einen Treffer im Gleichungssystem, wenn der entstandene
  // dynamische Faktor bereits in der dynamischen Faktorbasis bekannt ist.

  // Wenn ein neuer dynamischer Primfaktor entdeckt wird, so empfiehlt es sich,
  // in der Menge der Specialfaktoren (welches zusammengesetzte Faktoren sind) einen
  // Teilbarkeitstest auf diesen Primfaktor durchzufhren. Auf diese Weise knnen
  // weitere Primfaktoren isoliert werden, die als dynamische Faktoren zur Verfgung stehen.
  // Im Gegensatz zur bisherigen BUSY-Methode verwandelt die LAZY-Methode
  // den abgesplitteten Primfaktor in einen Korrekturfaktor. Die eigentliche
  // Auswertung der Relation wird damit auf einen spteren Zeitpunkt verschoben
  // und mglicherweise ganz verhindert, falls die betroffene Relation nicht
  // bentigt wird.

  static int Trashcounter = 0; /* Die (nach Elimination eines Faktors) nutzlos auf dem File
				  befindlichen Relationen */
  
  TSpecialFactorRelation SearchRelation;
  SearchRelation.factor.set_for_search(Primfaktor);

  while (true)
   {
    //cout << "splitten " << Primfaktor << endl;
    /* Wegen Rekursion etc. immer wieder korrekt neu aufsetzen, wenn
       ein Specialfaktor gekegelt wurde -> nchsthherer Special-Factor */
    const TSpecialFactorRelations::iterator pos=SpecialFactorRelations.lower_bound(SearchRelation);
    if (pos==SpecialFactorRelations.end()) break;
    if (!(*pos).factor.DLP_divisible_by(Primfaktor)) break;

    TSpecialFactorRelation gemerkte_Relation = *pos;
    SpecialFactorRelations.erase(pos); // remove old specialfactor
    gemerkte_Relation.factor.swap();
    SpecialFactorRelations.erase(gemerkte_Relation); // and remove also his swapped version
    gemerkte_Relation.factor.swap(); // swap back to original
    
    TDynamicFactorRelation relation;
    relation.factor=Primfaktor;
    const TDynamicFactorRelations::iterator p_el = DynamicFactorRelations.find(relation);
    if (p_el==DynamicFactorRelations.end()) 
      {
        MARK; 
        cerr << "Inconsistency in dynamic factorbase!" << endl;
        exit(1);
      }

    relation.factor=gemerkte_Relation.factor/Primfaktor; // dies ist der neue Restfaktor
    // now we have a correctly computed dynamic factor at hand

    // but is it really a new one, or did it exist already the dynamic factorbase?
    const TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(relation);
    if (p==DynamicFactorRelations.end())
     { 
       // er ist neu!
       //cout << "Special Factor: Dynamischer Faktor erzeugt!" << endl;
       statistical_data::Special_to_dynamic_Factor_hit++;
       relation.fpos = DynamicRelations_to_file.tellp();
       // nun die Relation modifiziert umkopieren:
       {
         const streampos rpos=SpecialRelations_from_file.tellg(); // Position merken
         SpecialRelations_from_file.seekg(gemerkte_Relation.fpos); // Position merken
         CmpqsFactor ret; string s;
         // Dateiformatflags retten
         const ios::fmtflags oldFlagsR = SpecialRelations_from_file.flags(); // alte Formatflags retten
         SpecialRelations_from_file.unsetf(ios::hex); SpecialRelations_from_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         const ios::fmtflags oldFlagsW = DynamicRelations_to_file.flags(); // alte Formatflags retten
         DynamicRelations_to_file.unsetf(ios::hex); DynamicRelations_to_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
         SpecialRelations_from_file >> s; // Startzeichen
         if (s != "G")
          {
	   MARK;
           cerr << "error: wrong position while reading relation" << endl;
           exit(1);
          }
         SpecialRelations_from_file >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
         SpecialRelations_from_file >> s; // delta
         DynamicRelations_to_file << "G " << setprecision(20) << relation.factor << " " << s << " ";

         // ab jetzt wird hexadezimal gearbeitet:
         SpecialRelations_from_file.unsetf(ios::dec); SpecialRelations_from_file.setf(ios::hex); // Hexadezimalausgabe
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         DynamicRelations_to_file.unsetf(ios::dec); DynamicRelations_to_file.setf(ios::hex); // Hexadezimalausgabe
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung

         int distance=GetValue(SpecialRelations_from_file);
         while (distance>=0)
          {
            PutValue(DynamicRelations_to_file,distance);
            distance=GetValue(SpecialRelations_from_file);
          }
         PutValue(DynamicRelations_to_file,-2);
         PutValue(DynamicRelations_to_file,Primfaktor);
         if (distance==-2)
          {
            distance=GetValue(SpecialRelations_from_file);
            while (distance>=0)
             {
               PutValue(DynamicRelations_to_file,distance);
               distance=GetValue(SpecialRelations_from_file);
             }            
          }
         PutValue(DynamicRelations_to_file,-1); DynamicRelations_to_file << endl;
         SpecialRelations_from_file.flags(oldFlagsR); // alte Formatflags wiederherstellen
         DynamicRelations_to_file.flags(oldFlagsW); // alte Formatflags wiederherstellen
         SpecialRelations_from_file.seekg(rpos); // position restaurieren
       }
       relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
#if !defined(IS_SERVER) && !defined(USE_FAKEHEAP)
       compute_Deltas_for(relation, SieveOffset+PhysicalSieveSize);
#endif
       DynamicFactorRelations.insert(relation); // Dynamicfactor einfgen
       SpecialRelations::split_by_primefactor(relation.factor);
     }
    else 
     { 
       // er existierte schon in der dyn. FB -> er fhrt zu einem Treffer im GLS!
       //cout << "Special Factor: Special Hit!" << endl;
       CRelation *GL = new CRelation;
       GL->combine(SpecialRelations_from_file,gemerkte_Relation.fpos);
       // gefundene dynamische Faktoren kombinieren:
       mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
       GL->combine(DynamicRelations_from_file,(*p).fpos);
       // nicht vergessen: auch Primfaktor eliminieren!
       mpz_mul_ui(GL->Delta,GL->Delta,(*p_el).factor); mpz_mod(GL->Delta,GL->Delta,n);
       GL->combine(DynamicRelations_from_file,(*p_el).fpos);
       statistical_data::Special_hit++; StaticRelations::insert(GL);
     } 

    { 
      // remove old (and superfluous) special-relation
      const streampos pos = SpecialRelations_to_file.tellp(); // save position
      SpecialRelations_to_file.seekp(gemerkte_Relation.fpos); SpecialRelations_to_file << "U "; // mark as invalid
      SpecialRelations_to_file.seekp(pos); // restore position
    }

    ++Trashcounter; // Die auf dem File verbleibenden sinnlosen Relationen zhlen
   }

 // Wenn sich sehr viel Mll angesammelt hat, knnte man nun eine Meldung
 // erstatten. Aber das drfte sich bei den heutigen Festplattengren
 // kaum lohnen; da nervt die Meldung mehr als sie ntzt...
}


bool SpecialRelations::insert(const CmpqsFactor &DLP, const string &GL_String)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren
  // und ins Gleichungssystem eintragen.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat! (in LAZY)" << endl;
     istringstream is(GL_String);
     CRelation* GL = new CRelation();
     GL->multi_combine_init();
     CmpqsFactor DLP_2 = GL->multi_combine_main(is);
     if (DLP!=DLP_2) { MARK; cerr << "DLP-Error somewhere in LAZY!" << endl; exit(1); }
     statistical_data::Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     StaticRelations::insert(GL,without_multi_combine_init); // LAZY -> einfgen geschieht hier...
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  const TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  if (pos==SpecialFactorRelations.end())
   {
     //cout << "Speichere DLP-Relation (LAZY)" << endl;
     //relation.fpos=GL->save(SpecialRelations_to_file,relation.factor);
     relation.fpos=SpecialRelations_to_file.tellp();
     SpecialRelations_to_file << GL_String;

     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
        SpecialRelations::CycleSearch();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit durch bekannten Special Factor! (LAZY)" << endl;
     istringstream is(GL_String);
     CRelation* GL = new CRelation();
     GL->multi_combine_init();
     CmpqsFactor DLP_2 = GL->multi_combine_main(is);
     if (DLP!=DLP_2) { MARK; cerr << "DLP-Error somewhere in LAZY!" << endl; exit(1); }
     statistical_data::Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.assign_to_mpz(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->multi_combine_main(SpecialRelations_from_file,(*pos).fpos);
     //GL->multi_combine_exit() kann entfallen, wenn zugehriges init bei StaticRelations::insert entfllt
     StaticRelations::insert(GL,without_multi_combine_init); // LAZY -> einfgen geschieht hier...
     return false;
   }
}

#endif /* IS_CLIENT not defined */
