/*! @file
 * @brief
 * contains some types and constants for qsieve.H and qsieve.cc
 */

#ifndef QSIEVE_FWD_HEADER_
#define QSIEVE_FWD_HEADER_

#include "utils.H"

const int server_port = 12364;
// server port for networked distributed factoring

const int XML_status_port = 12370;
// port for accessing XML status information


const int SingleLargePrime_Threshold = 1000000000;
// bis hierhin werden gefundene Faktoren als SingleLarge-Primes akzeptiert
// und in die Dynamic-Factor-Menge aufgenommen

const int DynamicFactor_SievingThreshold = MIN(SingleLargePrime_Threshold,100000000);
// bis hierhin wird mit den SingleLarge-Primes auch (aktiv) gesiebt.
// (Der Wert mu nicht notwendigerweise niedriger als SingleLargePrime_Threshold sein,
// aber wenn er kleiner ist, dann git es in den Siebroutinen fr Dynamische Faktoren
// einen CutOff an dieser Stelle.)


class StaticFactorbase
{
public:
 typedef unsigned short int FBsizetype; // Typ der Faktorbasisgre
  // Wenn die Faktorbasis maximal 65535 Elemente haben soll,
  // sollte man "unsigned short int" whlen, um Speicherplatz zu sparen.
  // Wenn Speicherplatz keine Rolle spielt, ist "int" eine gute Wahl (denn
  // dann gibt es u.U. weniger implizite Typkonvertierungen).
  // Hinweis:
  // "FBsizetype" hat *nur* mit der Anzahl der Primzahlen in der statischen
  // Faktorbasis (also den Indices auf die darin enthaltenen Primzahlen),
  // aber *nichts* mit dem Typ (und der Gre) der Primzahlen aus der
  // statischen Faktorbasis selbst zu tun!

protected:
 static FBsizetype Size_StaticFactorbase;

public:
 static const FBsizetype Size() { return Size_StaticFactorbase; }

 static const FBsizetype MaxSize = 30000; /* Maximale Gre der Primzahlbasis */ 
 /* Hinweis: Je grer die zu faktorisierende Zahl ist, desto grer
    sollte prinzipiell auch die Primzahlbasis sein, da dann die Relationen
    einfacher gefunden werden knnen.
    Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
    den Speicher!
    Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
    Primzahlbasis verlangsamend aus, da fr ein greres Gleichungssystem
    natrlich auch mehr Relationen gesiebt werden mssen.
    Je nach Geschwindigkeit des Rechners, des verfgbaren Haupt- und
    Hintergrundspeichers (und auch dessen Zugriffszeiten) knnen die
    ttschlich zu verwendenden Werte stark variieren. Sie sind deshalb
    experimentell zu ermitteln.
    Die Maximalgre wird durch MaxSize_StaticFactorbase festgelegt und steht
    damit bereits zur Kompilierzeit fest.
    Bei einer Speicherkapazitt von 32 MB (realem) Hauptspeicher
    hat sich 20000 als brauchbarer Wert erwiesen.
 */

 static const int max_additional_Powers = 1000; // maximale Anzahl weiterer Primzahlpotenzen, mit denen gesiebt wird

 friend void tune_parameters(const unsigned int);
};

#endif /*  QSIEVE_FWD_HEADER_ */
